module Main exposing (..)

import Browser
import Html exposing (..)
import Html.Attributes exposing (..)
import Html.Events exposing (..)
import Svg exposing (svg)
import Svg.Attributes
import Character exposing (Character)
import Sphere exposing (Sphere)
import BoundedInt exposing (BoundedInt)

main : Program () Model Msg
main = 
    Browser.document
        { init = init
        , view = view
        , update = update
        , subscriptions = subscriptions
        }

init : () -> ( Model, Cmd Msg )
init _ =
    (   { character = Character.new
        , modal = Nothing
        , modalValue = ""
        }
    , Cmd.none
    )

type alias Model = 
    { character : Character
    , modal : Maybe ModalType
    , modalValue : String
    }

type Msg 
    = OpenModal ModalType
    | CloseModal
    | ChangeModalValue String
    | SaveModalChange ModalType
    | ChangeSphere Sphere

type ModalType
    = Name

update : Msg -> Model -> ( Model, Cmd Msg )
update msg model = 
    case msg of
        OpenModal modalType -> 
            (   { model 
                | modal = ( Just modalType ) 
                , modalValue = modalValue modalType model.character
                }
            , Cmd.none
            )
        CloseModal ->
            ( { model | modal = Nothing }
            , Cmd.none
            )
        ChangeModalValue value ->
            ( { model | modalValue = value }
            , Cmd.none
            )
        SaveModalChange modalType ->
            case modalType of
                Name ->
                    (   { model 
                        | character = Character.changeName model.character model.modalValue 
                        , modal = Nothing
                        }
                    , Cmd.none
                    )
        ChangeSphere sphere ->
            (   { model
                | character = Character.changeSphere model.character sphere
                }
            , Cmd.none
            )

view : Model -> Browser.Document Msg
view model =
    { title = "Character Sheet"
    , body = 
        [ viewEditableText Name model.character.name
        , viewSpheres model.character.spheres
        , viewModal model
        ]
    }

viewEditableText : ModalType -> String -> Html Msg
viewEditableText modalType content =
    div []
        [ span [ class "icon-text" ]
            [ span [] [ text content ]
            , span 
                [ class "icon" 
                ] 
                [ i 
                    [ class "fas fa-pen"
                    , onClick ( OpenModal modalType ) 
                    ] []
                ]
            ]
        ]

viewSpheres : List Sphere -> Html Msg
viewSpheres spheres =
    div [] ( List.map viewSphere spheres )

viewSphere : Sphere -> Html Msg
viewSphere sphere =
    div []
        [ span [] [ text sphere.name ]
        , span [] [ viewDots sphere ]
        ]

viewModal : Model -> Html Msg
viewModal model =
    case model.modal of
        Nothing -> div [ class "modal" ] []
        Just modalType ->
            div [ class "modal is-active" ]
                [ div 
                    [ class "modal-background" 
                    , onClick CloseModal
                    ] []
                , div [ class "modal-card" ] 
                    [ header [ class "modal-card-head" ]
                        [ p [ class "modal-card-title" ] [ text ( "Change " ++ ( modalTitle modalType ) ) ]
                        ]
                    , section [ class "modal-card-body" ] 
                        [ p [] 
                            [ input 
                                [ class "input" 
                                , type_ "text"
                                , placeholder ( "Enter New " ++ ( modalTitle modalType ) )
                                , value model.modalValue
                                , onInput ( \x -> ChangeModalValue x )
                                ] []
                            ]
                        ]
                    , footer [ class "modal-card-foot" ] 
                        [ div [ class "buttons" ] 
                            [ button 
                                [ class "button is-success"
                                , onClick ( SaveModalChange modalType )
                                ] [ text "Save Changes" ]
                            , button 
                                [ class "button"
                                , onClick CloseModal
                                ] [ text "Cancel" ]
                            ]
                        ]
                    ]
                ]

modalTitle : ModalType -> String
modalTitle modalType =
    case modalType of
        Name -> "Character Name"

modalValue : ModalType -> Character -> String
modalValue modalType character =
    case modalType of
        Name -> character.name

viewDots : Sphere -> Html Msg
viewDots sphere =
    let
        dotSize = 12
        dotGap = 4
        dotOutline = 0.05 * dotSize
        marginalIncrease = ( dotSize + dotGap ) / 2
        svgOutlineMargin = ( sqrt ( 2 * ( dotOutline ^ 2 ) ) ) / 2

        rangeList = List.range 0 ( sphere.dots.max - 1 )

        isFilledList = rangeList |> List.map ( (>) sphere.dots.value )
    in
        svg 
            [ Svg.Attributes.width  ( String.fromFloat ( dotSize + marginalIncrease * toFloat ( sphere.dots.max - 1 ) + 2 * svgOutlineMargin ) )
            , Svg.Attributes.height ( String.fromFloat ( dotSize + marginalIncrease + 2 * svgOutlineMargin ) )
            ]
            ( rangeList
                |> List.map ( toDotPosition dotSize dotGap svgOutlineMargin )
                |> List.map3 ( viewDot sphere dotSize dotOutline ) isFilledList rangeList
            )

toDotPosition : Float -> Float -> Float -> Int -> ( Float, Float )
toDotPosition size gap offset id =
    let
        marginalIncrease = ( size + gap ) / 2
    in
        ( toFloat id * marginalIncrease + offset
        , toFloat ( remainderBy 2 ( id + 1 ) ) * marginalIncrease + offset
        )

viewDot : Sphere -> Float -> Float -> Bool -> Int -> ( Float, Float ) -> Html Msg
viewDot sphere size strokeWidth isFilled id position =
    Svg.polygon 
        [ Svg.Attributes.points 
            (   [ ( 0.5, 0.0 )
                , ( 1.0, 0.5 )
                , ( 0.5, 1.0 )
                , ( 0.0, 0.5 )
                ]
                |> List.map ( Tuple.mapBoth   ( (*) size ) ( (*) size ) )
                |> List.map ( Tuple.mapFirst  ( (+) ( Tuple.first  position ) ) )
                |> List.map ( Tuple.mapSecond ( (+) ( Tuple.second position ) ) )
                |> List.map tupleToString
                |> List.intersperse " "
                |> String.concat
            )
        , Svg.Attributes.fill "var(--bulma-body-color)"
        , Svg.Attributes.fillOpacity ( boolAsIntString isFilled )
        , Svg.Attributes.stroke "var(--bulma-body-color)"
        , Svg.Attributes.strokeWidth ( String.fromFloat strokeWidth )
        , onClick ( ChangeSphere { sphere | dots = BoundedInt.change sphere.dots ( valueOrReset sphere.dots ( id + 1 ) ) } )
        ] []

tupleToString : ( Float, Float ) -> String
tupleToString tuple =
    let
        first  = String.fromFloat ( Tuple.first  tuple )
        second = String.fromFloat ( Tuple.second tuple )
    in
        first ++ "," ++ second

boolAsIntString : Bool -> String
boolAsIntString bool =
    if bool
    then "1"
    else "0"

valueOrReset : BoundedInt -> Int -> Int
valueOrReset dots value =
    if dots.value == value
    then dots.min
    else value

subscriptions : Model -> Sub Msg
subscriptions _ =
  Sub.none

ariaLabel : String -> Attribute msg
ariaLabel value =
    Html.Attributes.attribute "aria-label" value